﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Data.SqlClient;
using System.Data.SqlServerCe;
using System.Linq;
using System.Security.Cryptography.X509Certificates;
using HIPS.CommonSchemas;
using HIPS.CommonSchemas.PatientIdentifier;
using HIPS.PcehrDataStore.DataAccess;
using HIPS.PcehrDataStore.Schemas;
using HIPS.PcehrDataStore.Schemas.Enumerators;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Test.Helpers;

namespace Test.MedicareCCA.Helpers
{
    /// <summary>
    /// Provides patients for use in CCA testing.
    /// </summary>
    public class CcaPatient : PatientTestHelpersBase
    {
        #region Configuration Properties

        /// <summary>
        /// The patient identifier type.
        /// </summary>
        public static PatientIdentifierType PatientIdentifierType
        {
            get
            {
                return (PatientIdentifierType)Enum.Parse(typeof(PatientIdentifierType), ConfigurationManager.AppSettings["PatientIdentifierType"].ToString());
            }
        }

        /// <summary>
        /// Which of the rows in the CcaTestPatients table TestPcehr has the data for this patient.
        /// </summary>
        public string TestDataId
        {
            get;
            set;
        }

        private SqlCeConnection ccaTestPatientsConnectionField;

        private SqlCeConnection CcaTestPatientsConnection
        {
            get
            {
                if (ccaTestPatientsConnectionField == null)
                {
                    string connectionString = ConfigurationManager.ConnectionStrings["CcaTestPatientsConnectionString"].ConnectionString;
                    ccaTestPatientsConnectionField = new SqlCeConnection(connectionString);
                    ccaTestPatientsConnectionField.Open();
                }
                return ccaTestPatientsConnectionField;
            }
        }

        private SqlCeCommand GetCcaTestPatientsCommand()
        {
            return CcaTestPatientsConnection.CreateCommand();
        }

        private Dictionary<string, string> testPcehrField;

        private Dictionary<string, string> TestPcehr
        {
            get
            {
                if (testPcehrField == null)
                {
                    testPcehrField = new Dictionary<string, string>();
                    using (SqlCeCommand c = GetCcaTestPatientsCommand())
                    {
                        c.CommandText = string.Format("SELECT * FROM TestPcehr WHERE TestDataId='{0}'", TestDataId);
                        SqlCeResultSet resultSet = c.ExecuteResultSet(ResultSetOptions.None);
                        bool gotRow = resultSet.Read();
                        Assert.IsTrue(gotRow, DialogueResource.TestDataIdNotFound, TestDataId);
                        for (int i = 0; i < resultSet.FieldCount; i++)
                        {
                            testPcehrField[resultSet.GetName(i)] = resultSet.GetString(i);
                        }
                    }
                }
                return testPcehrField;
            }
        }

        /// <summary>
        /// The family name for the test patient.
        /// </summary>
        private string TestPatientFamilyName
        {
            get
            {
                if (TestDataId == null)
                {
                    return ConfigurationManager.AppSettings["TestPatientFamilyName"].ToString();
                }
                else
                {
                    return TestPcehr["FamilyName"];
                }
            }
        }

        /// <summary>
        /// The given name for the test patient.
        /// </summary>
        private string TestPatientGivenName
        {
            get
            {
                if (TestDataId == null)
                {
                    return ConfigurationManager.AppSettings["TestPatientGivenName"].ToString();
                }
                else
                {
                    return TestPcehr["FirstGivenName"];
                }
            }
        }

        /// <summary>
        /// The date of birth for the test patient.
        /// </summary>
        public DateTime TestPatientDateOfBirth
        {
            get
            {
                if (TestDataId == null)
                {
                    return DateTime.Parse(ConfigurationManager.AppSettings["TestPatientDateOfBirth"].ToString());
                }
                else
                {
                    return DateTime.Parse(TestPcehr["DateOfBirth"]);
                }
            }
        }

        /// <summary>
        /// The sex for the test patient.
        /// </summary>
        private SexEnumerator TestPatientSex
        {
            get
            {
                if (TestDataId == null)
                {
                    return (SexEnumerator)Enum.Parse(typeof(SexEnumerator), ConfigurationManager.AppSettings["TestPatientSex"].ToString());
                }
                else
                {
                    switch (TestPcehr["Sex"])
                    {
                        case "M": return SexEnumerator.Male;
                        case "F": return SexEnumerator.Female;
                        case "I": return SexEnumerator.IntersexOrIndeterminate;
                        case "N": return SexEnumerator.NotStatedOrInadequatelyDescribed;
                        default: throw new InvalidOperationException("Invalid Sex in TestPcehr table");
                    }
                }
            }
        }

        /// <summary>
        /// The valid IHI for the test patient.
        /// </summary>
        public string TestPatientValidIhi
        {
            get
            {
                if (TestDataId == null)
                {
                    return ConfigurationManager.AppSettings["TestPatientValidIhi"].ToString();
                }
                else
                {
                    return TestPcehr["Ihi"];
                }
            }
        }

        /// <summary>
        /// An invalid IHI for the test patient.
        /// </summary>
        public string TestPatientInvalidIhi
        {
            get
            {
                if (TestDataId == null)
                {
                    return ConfigurationManager.AppSettings["TestPatientInvalidIhi"].ToString();
                }
                else
                {
                    // Add one to the 15-character prefix of the valid IHI and recalculate the check digit
                    string fifteenDigit = (long.Parse(TestPcehr["Ihi"].Substring(0, 15)) + 1).ToString();
                    return Nehta.VendorLibrary.Common.Validation.CalculateLuhnCheckDigit(fifteenDigit);
                }
            }
        }

        /// <summary>
        /// The hospital to use for CCA testing, specified in the app.config file for the CCA test project.
        /// </summary>
        private static int TestHospitalId
        {
            get
            {
                return int.Parse(ConfigurationManager.AppSettings["TestHospitalId"].ToString());
            }
        }

        /// <summary>
        /// The Patient Master ID will be inserted into the placeholder in this format to generate an MRN for the test patient.
        /// </summary>
        private static string TestPatientMrnFormat
        {
            get
            {
                return ConfigurationManager.AppSettings["TestPatientMrnFormat"].ToString();
            }
        }

        /// <summary>
        /// The number of days old to make the IHI when it should be outside the validation period, specified in the app.config file for the CCA test project.
        /// </summary>
        private static int OutsideValidationPeriodDays
        {
            get
            {
                return int.Parse(ConfigurationManager.AppSettings["OutsideValidationPeriodDays"].ToString());
            }
        }

        /// <summary>
        /// The Record Access Code (RAC, aka PACC) for the test patient.
        /// </summary>
        public string TestPatientRecordAccessCode
        {
            get
            {
                if (TestDataId == null)
                {
                    return ConfigurationManager.AppSettings["TestPatientRecordAccessCode"].ToString();
                }
                else
                {
                    return TestPcehr["RecordCode"];
                }
            }
        }

        /// <summary>
        /// The Limited Document Access Code (LDAC, aka PACCX) for the test patient.
        /// </summary>
        public string TestPatientLimitedDocumentAccessCode
        {
            get
            {
                if (TestDataId == null)
                {
                    return ConfigurationManager.AppSettings["TestPatientLimitedDocumentAccessCode"].ToString();
                }
                else
                {
                    return TestPcehr["DocumentCode"];
                }
            }
        }

        /// <summary>
        /// The path to a CDA document file to upload.
        /// </summary>
        public static string TestDocumentPath
        {
            get
            {
                return ConfigurationManager.AppSettings["TestDocumentPath"].ToString();
            }
        }

        /// <summary>
        /// The The Patient Master ID will be inserted into the "{0}" placeholder in
        /// this format to generate a source system episode id (visit number) for the
        /// test patient.
        /// </summary>
        public static string TestPatientEpisodeIdFormat
        {
            get
            {
                return ConfigurationManager.AppSettings["TestPatientEpisodeIdFormat"].ToString();
            }
        }

        /// <summary>
        /// Number of seconds to wait for queued operation to change from Pending status.
        /// </summary>
        public static int PcehrQueueTimeout
        {
            get
            {
                return int.Parse(ConfigurationManager.AppSettings["PcehrQueueTimeout"].ToString());
            }
        }

        #endregion Configuration Properties

        #region Test Targets

        // Note, the base class provides the TargetPatient property.

        /// <summary>
        /// The hospital patient to use for testing
        /// </summary>
        public HospitalPatient TargetHospitalPatient
        {
            get
            {
                if (testHospitalPatientField == null)
                {
                    string mrn = string.Format(TestPatientMrnFormat, TargetPatient.PatientMasterId);
                    testHospitalPatientField = InsertHospitalPatient(mrn);
                }
                return testHospitalPatientField;
            }
        }

        /// <summary>
        /// The hospital to use for testing.
        /// </summary>
        public Hospital TargetHospital
        {
            get
            {
                if (testHospitalField == null)
                {
                    ReloadTargetHospital();
                }
                return testHospitalField;
            }
        }

        /// <summary>
        /// The episode to use for testing.
        ///
        /// If the patient identifier type is ValidatedIhi then the episode is not actually inserted into the database by this method.
        /// </summary>
        public Episode TargetEpisode
        {
            get
            {
                if (testEpisodeField == null)
                {
                    testEpisodeField = InsertEpisode();
                }
                return testEpisodeField;
            }
        }

        /// <summary>
        /// The code system of the code to use for this hospital. This value
        /// should exist in CodeSystem.Code. There should be a row in
        /// HospitalCode that links the CodeSystemId of the row in CodeSystem
        /// to the HospitalId that was specified above, and contains a code
        /// for the hospital.
        /// </summary>
        private string HospitalCodeSystem
        {
            get
            {
                return ConfigurationManager.AppSettings["HospitalCodeSystem"].ToString();
            }
        }

        /// <summary>
        /// When looking for the episode that HIPS created for a call with a ValidatedIhi,
        /// it will be matched by admission date/time with a tolerance set here.
        /// </summary>
        private double MaximumAdmissionTimeDifferenceSeconds
        {
            get
            {
                return double.Parse(ConfigurationManager.AppSettings["MaximumAdmissionTimeDifferenceSeconds"].ToString());
            }
        }

        private string targetHospitalCode = null;

        /// <summary>
        /// Gets the hospital code.
        /// </summary>
        private string TargetHospitalCode
        {
            get
            {
                if (targetHospitalCode == null)
                {
                    List<HospitalCode> codes = new HIPS.PcehrDataStore.DataAccess.HospitalCodeDl().GetAll(TestHospitalId);
                    HospitalCode theCode = codes.FirstOrDefault(a => a.CodeSystemCode == HospitalCodeSystem);
                    if (theCode == null)
                    {
                        Assert.Fail("The hospital {0} does not have a code of type {1}", TestHospitalId, HospitalCodeSystem);
                    }
                    targetHospitalCode = theCode.Code;
                }
                return targetHospitalCode;
            }
        }

        /// <summary>
        /// Returns a patient identifier that can be sent to HIPS to identify the test patient and hospital.
        /// </summary>
        public PatientIdentifierBase TargetPatientIdentifier
        {
            get
            {
                switch (PatientIdentifierType)
                {
                    case Helpers.PatientIdentifierType.Mrn:
                        return new Mrn(TargetHospitalPatient.Mrn, TargetHospitalCode, HospitalCodeSystem);

                    case Helpers.PatientIdentifierType.PatientMasterId:
                        return new PatientMasterId(TargetPatient.PatientMasterId.Value, null, null, TestHospitalId);

                    case Helpers.PatientIdentifierType.StatePatientId:
                        return new StatePatientId(TargetPatient.StatePatientId, TargetHospitalCode, HospitalCodeSystem);

                    case Helpers.PatientIdentifierType.ValidatedIhi:
                        return new ValidatedIhi(
                            IhiInformation.Ihi,
                            (IhiStatus)IhiInformation.IhiStatusId,
                            (IhiRecordStatus)IhiInformation.IhiRecordStatusId,
                            IhiInformation.DateLastValidated,
                            TestPatientFamilyName,
                            TestPatientGivenName,
                            TestPatientDateOfBirth,
                            TestPatientSex,
                            TargetHospitalCode,
                            HospitalCodeSystem);
                }
                throw new InvalidOperationException();
            }
        }

        /// <summary>
        /// Returns a patient demographic record.
        /// </summary>
        public Demographic TargetDemographic
        {
            get
            {
                return new Demographic(
                    TestPatientFamilyName,
                    TestPatientGivenName,
                    TestPatientDateOfBirth,
                    TestPatientSex,
                    TestPcehr["MedicareCardNumber"],
                    TestPcehr["MedicareIrn"],
                    TargetHospitalCode,
                    HospitalCodeSystem);
            }
        }

        /// <summary>
        /// Obtains the target patient master ID. If the target patient master was inserted
        /// by HIPS rather than by the test suite, then this involves a query on the database.
        /// </summary>
        public int TargetPatientMasterId
        {
            get
            {
                if (!TargetPatient.PatientMasterId.HasValue)
                {
                    GetDatabaseIds(needEpisode: false);
                }
                return TargetPatient.PatientMasterId.Value;
            }
        }

        public PatientMasterIhi IhiInformation
        {
            get;
            set;
        }

        #endregion Test Targets

        #region Public Methods

        /// <summary>
        /// Sets up a patient ready for making a request to the PCEHR. Deletes
        /// the test patient (including all child data, audit records) if it
        /// already exists.
        ///
        /// If the patient identifier type is not ValidatedIhi then the patient,
        /// IHI information and hospital patient will be pre-populated into the
        /// database as if they had come in via the PAS feed.
        ///
        /// However, if the patient identifier type is ValidatedIhi then nothing
        /// will be added to the database, and the IHI information will merely
        /// be stored in the property IhiInformation, for use when you retrieve
        /// the patient identifier using TargetPatientIdentifier.
        /// </summary>
        /// <param name="ihiValid"></param>
        /// <param name="validatedWithinPeriod"></param>
        /// <param name="hasUnresolvedAlerts"></param>
        /// <returns></returns>
        public static CcaPatient GetPatient(bool ihiValid, bool validatedWithinPeriod, bool hasUnresolvedAlerts, string testDataId = null, bool noihi = false)
        {
            CcaPatient item = new CcaPatient();
            item.TestDataId = testDataId;
            item.ResetPatientData();
            item.IhiInformation = new PatientMasterIhi();
            if (PatientIdentifierType != Helpers.PatientIdentifierType.ValidatedIhi)
            {
                item.IhiInformation.PatientMasterId = item.SetTestData();
            }

            if (ihiValid)
            {
                item.IhiInformation.Ihi = item.TestPatientValidIhi;
                item.IhiInformation.IhiStatusId = (int)IhiStatus.Active;
                item.IhiInformation.IhiRecordStatusId = (int)IhiRecordStatus.Verified;
            }
            else
            {
                item.IhiInformation.Ihi = item.TestPatientInvalidIhi;
                item.IhiInformation.IhiStatusId = (int)IhiStatus.Active;
                item.IhiInformation.IhiRecordStatusId = (int)IhiRecordStatus.Verified;
            }

            if (noihi)
            {
                item.IhiInformation.Ihi = null;
                item.IhiInformation.IhiStatusId = (int)IhiStatus.Unknown;
                item.IhiInformation.IhiRecordStatusId = (int)IhiRecordStatus.Unknown;
            }

            if (hasUnresolvedAlerts)
            {
                item.IhiInformation.IhiStatusId = (int)IhiStatus.DuplicateIhi;
            }

            item.IhiInformation.RegisteredFamilyName = item.TargetPatient.CurrentName.FamilyName;
            item.IhiInformation.RegisteredGivenName = item.TargetPatient.CurrentName.GivenNames;
            item.IhiInformation.RegisteredSexId = item.TargetPatient.CurrentSexId;
            if (validatedWithinPeriod)
            {
                item.IhiInformation.DateLastValidated = DateTime.Now;
            }
            else
            {
                item.IhiInformation.DateLastValidated = DateTime.Now - TimeSpan.FromDays(OutsideValidationPeriodDays);
            }

            if (PatientIdentifierType != Helpers.PatientIdentifierType.ValidatedIhi)
            {
                item.PatientMasterIhiInsert(item.IhiInformation);
                HospitalPatient hospitalPatient = item.TargetHospitalPatient;
            }
            return item;
        }

        /// <summary>
        /// Gets a test patient without an IHI.
        /// </summary>
        /// <param name="testDataId">The test data ID from the SQL compact database (or null for default)</param>
        /// <returns>The CCA patient</returns>
        public static CcaPatient GetPatientWithoutIhi(string testDataId = null)
        {
            CcaPatient item = new CcaPatient();
            item.TestDataId = testDataId;
            item.ResetPatientData();
            item.IhiInformation = new PatientMasterIhi();
            item.IhiInformation.IhiStatusId = (int)IhiStatus.Unknown;
            item.IhiInformation.IhiRecordStatusId = (int)IhiStatus.Unknown;
            item.IhiInformation.DateLastValidated = DateTime.Now;
            if (PatientIdentifierType != Helpers.PatientIdentifierType.ValidatedIhi)
            {
                item.SetTestData();
                HospitalPatient hospitalPatient = item.TargetHospitalPatient;
            }
            return item;
        }

        /// <summary>
        /// Resets the date last validated for the IHI to be outside the validation period.
        /// </summary>
        public void MakeIhiStale()
        {
            IhiInformation.DateLastValidated = DateTime.Now - TimeSpan.FromDays(OutsideValidationPeriodDays);

            if (PatientIdentifierType != Helpers.PatientIdentifierType.ValidatedIhi)
            {
                PatientMasterIhiDelete(IhiInformation);
                PatientMasterIhiInsert(IhiInformation);
            }
        }

        /// <summary>
        /// Resets the IHI to be valid.
        /// </summary>
        public void MakeIhiValid()
        {
            IhiInformation.Ihi = TestPatientValidIhi;
            IhiInformation.IhiRecordStatusId = (int)IhiRecordStatus.Verified;

            if (PatientIdentifierType != Helpers.PatientIdentifierType.ValidatedIhi)
            {
                PatientMasterIhiDelete(IhiInformation);
                PatientMasterIhiInsert(IhiInformation);
            }
        }

        /// <summary>
        /// Resets the IHI to be invalid.
        /// </summary>
        public void MakeIhiInvalid()
        {
            IhiInformation.Ihi = TestPatientInvalidIhi;
            IhiInformation.IhiRecordStatusId = (int)IhiRecordStatus.Unknown;

            if (PatientIdentifierType != Helpers.PatientIdentifierType.ValidatedIhi)
            {
                PatientMasterIhiDelete(IhiInformation);
                PatientMasterIhiInsert(IhiInformation);
            }
        }

        /// <summary>
        /// Gets all the downloaded documents for the target patient.
        /// </summary>
        /// <returns>A list of downloaded document.</returns>
        public List<DownloadedDocument> GetAllDownloadedDocuments()
        {
            if (!TargetPatient.PatientMasterId.HasValue)
            {
                GetDatabaseIds(needEpisode: false);
            }
            DownloadedDocumentDl dataAccess = new DownloadedDocumentDl(GetTestUser());
            return dataAccess.GetAll(TargetPatient.PatientMasterId.Value);
        }

        /// <summary>
        /// Gets the PCEHR message queue items for the target episode.
        /// </summary>
        /// <param name="episodeId">The episode ID</param>
        /// <returns></returns>
        public List<PcehrMessageQueue> GetPcehrMessageQueueItems()
        {
            if (!TargetEpisode.EpisodeId.HasValue)
            {
                GetDatabaseIds(needEpisode: true);
            }
            GenericDataAccess dataAccess = new GenericDataAccess();
            List<PcehrMessageQueue> result;
            using (SqlCommand command = TestCommand)
            {
                command.CommandText = string.Format("SELECT * from [hips].[PcehrMessageQueue] WHERE EpisodeId = {0} ORDER BY PcehrMessageQueueId DESC", TargetEpisode.EpisodeId);
                result = GetPopulatedBusinessList<PcehrMessageQueue>(command.ExecuteReader());
            }
            return result;
        }

        public void ReloadTargetHospital()
        {
            testHospitalField = new HospitalDl().Get(TestHospitalId);
        }

        #endregion Public Methods

        #region Private Implementation

        private HospitalPatient testHospitalPatientField;
        private Episode testEpisodeField;
        private Hospital testHospitalField;

        /// <summary>
        /// The constructor is private. Use GetPatient to allocate an instance of CcaPatient.
        /// </summary>
        private CcaPatient()
        {
        }

        /// <summary>
        /// Returns a SQL command to match the target patient in the database,
        /// so that the test suite can remove all records relating to that
        /// patient, and start each test from a clean state.
        /// </summary>
        /// <returns></returns>
        public override string SqlCommandText()
        {
            return string.Format(
                @"SELECT DISTINCT pm.PatientMasterId FROM hips.PatientMaster pm
                INNER JOIN hips.PatientMasterName pmn on pm.PatientMasterId = pmn.PatientMasterId
                WHERE DateOfBirth = '{0:yyyy-MM-dd}'
                AND FamilyName = '{1}'
                AND GivenNames = '{2}'
                ORDER BY pm.PatientMasterId DESC",
                TestPatientDateOfBirth,
                TestPatientFamilyName,
                TestPatientGivenName);
        }

        /// <summary>
        /// When calling HIPS with a ValidatedIhi, the database primary keys for the
        /// patient master, hospital patient and episode are allocated by HIPS and
        /// not by this test suite. Therefore we will query to find them from the
        /// database, using the same query as was used to find and remove the patient.
        /// </summary>
        internal void GetDatabaseIds(bool needEpisode)
        {
            // Get the Patient Master ID
            GenericDataAccess dataAccess = new GenericDataAccess();
            using (SqlCommand command = TestCommand)
            {
                command.CommandText = SqlCommandText();
                int? result = System.Convert.ToInt32(command.ExecuteScalar());
                if (result.HasValue && result > 0)
                {
                    TargetPatient.PatientMasterId = result;
                }
                command.Connection.Close();
            }
            if (!TargetPatient.PatientMasterId.HasValue)
            {
                Assert.Fail(DialogueResource.PatientNotFound);
            }

            // Get the Hospital Patient and Episode
            UserDetails user = GetTestUser();
            HospitalPatient hospitalPatient = new HospitalPatientDl(user).GetAll(TestHospitalId, TargetPatient.PatientMasterId.Value).First();

            //All Episode LifeCycle Types can be used in this case
            var matched = from episode in new EpisodeDl(user).GetAll(hospitalPatient.PatientId.Value, null)
                          where Math.Abs((episode.AdmissionDate - TargetEpisode.AdmissionDate).TotalMinutes) < MaximumAdmissionTimeDifferenceSeconds
                          select episode;
            if (matched.Count() == 1)
            {
                Episode episode = matched.First();
                TargetEpisode.EpisodeId = episode.EpisodeId.Value;
            }
            else if (needEpisode)
            {
                Assert.Fail(DialogueResource.EpisodeNotFound);
            }
            TargetHospitalPatient.PatientId = hospitalPatient.PatientId.Value;
        }

        /// <summary>
        /// Overrides the basic patient with the one we want to use for CCA testing.
        /// </summary>
        override public void GetTestInsertPatient()
        {
            base.GetTestInsertPatient();
            TargetPatient.Names = new List<PatientMasterName>();
            TargetPatient.SetNewCurrentName(null, TestPatientGivenName, TestPatientFamilyName, null);
            TargetPatient.CurrentSexId = (int)TestPatientSex;
            TargetPatient.DateOfBirth = TestPatientDateOfBirth;
        }

        /// <summary>
        /// Creates an episode for the test patient and returns the episode that was created.
        ///
        /// If the patient identifier type is ValidatedIhi then the episode is not actually inserted into the database by this method.
        /// </summary>
        /// <returns>The episode that was created</returns>
        private Episode InsertEpisode()
        {
            Episode item = new Episode();
            item.AdmissionDate = DateTime.Now;
            item.SourceSystemEpisodeId = string.Format(TestPatientEpisodeIdFormat, TargetHospitalPatient.PatientMasterId);
            item.ResponsibleProviderId = -1;
            item.EpisodeLifecycleId = (int)HIPS.PcehrDataStore.Schemas.Enumerators.EpisodeLifecycle.Admitted;
            item.EpisodeTypeId = -1;

            if (PatientIdentifierType != Helpers.PatientIdentifierType.ValidatedIhi)
            {
                item.PatientId = TargetHospitalPatient.PatientId.Value;
                GenericDataAccess dataAccess = new GenericDataAccess();
                using (SqlCommand command = TestCommand)
                {
                    command.CommandText = "[hips].[EpisodeInsert]";
                    command.CommandType = System.Data.CommandType.StoredProcedure;
                    Insert<Episode>(item, command);
                    command.CommandType = System.Data.CommandType.Text;
                }
            }
            return item;
        }

        /// <summary>
        /// Creates a hospital patient record with the specified MRN.
        /// If the patient identifier type is Validated IHI then the record
        /// will not be saved to the database.
        /// </summary>
        /// <param name="item">The item.</param>
        private HospitalPatient InsertHospitalPatient(string mrn)
        {
            HospitalPatient item = new HospitalPatient();
            item.HospitalId = TestHospitalId;
            item.Mrn = mrn;

            if (PatientIdentifierType != Helpers.PatientIdentifierType.ValidatedIhi)
            {
                item.PatientMasterId = TargetPatient.PatientMasterId.Value;
                GenericDataAccess dataAccess = new GenericDataAccess();
                using (SqlCommand command = TestCommand)
                {
                    command.CommandText = "[hips].[HospitalPatientInsert]";
                    command.CommandType = System.Data.CommandType.StoredProcedure;
                    Insert<HospitalPatient>(item, command);
                    command.CommandType = System.Data.CommandType.Text;
                }
            }
            return item;
        }

        /// <summary>
        /// Gets all the IHI lookup audit records for the target patient.
        /// </summary>
        /// <returns></returns>
        internal List<IhiLookupAudit> GetIhiLookupAudits()
        {
            int patientMasterId = TargetPatientMasterId;
            GenericDataAccess dataAccess = new GenericDataAccess();
            List<IhiLookupAudit> result;
            using (SqlCommand command = TestCommand)
            {
                command.CommandText = string.Format("SELECT * from [hips].[IhiLookupAudit] WHERE PatientMasterId = {0} ORDER BY IhiLookupAuditId", patientMasterId);
                result = GetPopulatedBusinessList<IhiLookupAudit>(command.ExecuteReader());
            }
            return result;
        }

        /// <summary>
        /// Gets all the PCEHR audit records for the target patient.
        /// </summary>
        /// <returns></returns>
        internal List<PcehrAudit> GetPcehrAudits()
        {
            int patientMasterId = TargetPatientMasterId;
            GenericDataAccess dataAccess = new GenericDataAccess();
            List<PcehrAudit> result;
            using (SqlCommand command = TestCommand)
            {
                command.CommandText = string.Format("SELECT * from [hips].[PcehrAudit] WHERE PatientMasterID = {0} ORDER BY PcehrAuditId", patientMasterId);
                result = GetPopulatedBusinessList<PcehrAudit>(command.ExecuteReader());
            }
            return result;
        }

        /// <summary>
        /// Gets all the consent audit records for the target episode.
        /// </summary>
        /// <returns></returns>
        internal List<ConsentAudit> GetConsentAudits()
        {
            if (!TargetEpisode.EpisodeId.HasValue)
            {
                GetDatabaseIds(needEpisode: true);
            }
            int episodeId = TargetEpisode.EpisodeId.Value;
            GenericDataAccess dataAccess = new GenericDataAccess();
            List<ConsentAudit> result;
            using (SqlCommand command = TestCommand)
            {
                command.CommandText = string.Format("SELECT * from [hips].[ConsentAudit] WHERE EpisodeId = {0} ORDER BY ConsentAuditId", episodeId);
                result = GetPopulatedBusinessList<ConsentAudit>(command.ExecuteReader());
            }
            return result;
        }

        /// <summary>
        /// Gets the signing certificate from the Windows certificate repository.
        /// </summary>
        /// <returns>Matching certificate.</returns>
        public X509Certificate2 GetSigningCertificate()
        {
            StoreLocation storeLocation = StoreLocation.CurrentUser;
            StoreName storeName = StoreName.My;
            X509FindType findType = X509FindType.FindBySerialNumber;
            HospitalDl hospitalDl = new HospitalDl();
            string findValue = hospitalDl.Get(TargetHospitalPatient.HospitalId).PcehrCertSerial;
            bool valid = false;

            X509Store certStore = new X509Store(storeName, storeLocation);
            certStore.Open(OpenFlags.ReadOnly);

            X509Certificate2Collection foundCerts =
              certStore.Certificates.Find(findType, findValue, valid);
            certStore.Close();

            // Check if any certificates were found with the criteria
            if (foundCerts.Count == 0)
                throw new ArgumentException("Certificate was not found with criteria '" + findValue + "'");

            // Check if more than one certificate was found with the criteria
            if (foundCerts.Count > 1)
                throw new ArgumentException("More than one certificate found with criteria '" + findValue + "'");

            return foundCerts[0];
        }

        #endregion Private Implementation
    }
}