﻿using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;
using System.Xml;
using System.Xml.Schema;
using HIPS.CommonSchemas;
using HIPS.CommonSchemas.PatientIdentifier;
using HIPS.PcehrDataStore.Schemas;
using HIPS.PcehrSchemas;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Test.CommonCcaNoc.Helpers;

namespace Test.NpdrCcaNoc.NoticeOfConnection
{
    /// <summary>
    /// Conformance Test Specification: NPDR Notice of Connection Test Plan
    ///                                 Version 1.1 — 15 July 2013
    /// Operation:  uploadDocument
    /// </summary>
    [TestClass]
    public class NOC_UploadDocument : CcaTest
    {
        /// <summary>
        /// Test No:         NOC_DC_001
        /// 
        /// Objective/Input:
        /// Perform a Provide and Register Document set - b operation with
        /// valid input elements for a NPDR Prescription Record
        ///
        /// Expected Result:
        /// Document will be stored and registered in the NPDR and will
        /// successfully be registered by the PCEHR.
        /// </summary>
        [TestMethod]
        [TestCategory("NPDR_NOC")]
        public void NPDR_NOC_01()
        {
            TestSuccessfulUpload(SampleDocumentType.PcehrPrescriptionRecord);
        }

        /// <summary>
        /// Test No:         NOC_DC_002
        /// 
        /// Objective/Input:
        /// Perform a Provide and Register Document set - b operation with
        /// valid input elements for a NPDR Dispense Record
        ///
        /// Expected Result:
        /// Document will be stored and registered in the NPDR and will
        /// successfully be registered by the PCEHR.
        /// </summary>
        [TestMethod]
        [TestCategory("NPDR_NOC")]
        public void NPDR_NOC_02()
        {
            TestSuccessfulUpload(SampleDocumentType.PcehrDispenseRecord);
        }
        
        /// <summary>
        /// Tests the successful upload of a document to NPDR.
        /// </summary>
        /// <param name="docType"></param>
        private void TestSuccessfulUpload(SampleDocumentType docType)
        {
            patient = CcaPatient.GetPatient(ihiValid: true, validatedWithinPeriod: true, hasUnresolvedAlerts: false, testDataId: "NOC ID 17");
            UserDetails user = patient.GetTestUser();
            PatientIdentifierBase identifier = patient.TargetPatientIdentifier;
            CdaDocument document = patient.CreateNewDocument(docType);
            string documentFormatCode = "1.2.36.1.2001.1006.1.170.2";
            HipsResponse hipsResponse = ProxyHelper.PcehrProxy.UploadOrSupersedeDocument(document.GetBytes(), identifier, user, null, patient.TargetEpisode.AdmissionDate, documentFormatCode);
            LogAssert.ExpectResponse(HipsResponseIndicator.OK, hipsResponse, DialogueResource.HipsServiceUploadDocument);
            QueueHelper.WaitForQueuedOperation(patient, LogAssert, DialogueResource.UploadWaitTimeout);

            // Get the Audit
            PcehrAudit audit = patient.GetLastPcehrAudit();
            LogAssert.AreEqual(patient.TargetPatientMasterId, audit.PatientMasterId, DialogueResource.PcehrAuditPatientMasterId);
            LogAssert.AreEqual(AuditOperationNames.UploadDocument, audit.ServiceName, DialogueResource.PcehrAuditServiceName);

            // Confirm that the payload is well formed and complies with the doesPCEHRExist Request XML Schema
            XmlDocument soapRequest = new XmlDocument();
            using (StringReader sr = new StringReader(audit.Request))
            {
                soapRequest.Load(sr);
            }
            XmlNamespaceManager nsmgr = new XmlNamespaceManager(new NameTable());
            nsmgr.AddNamespace("s", "http://www.w3.org/2003/05/soap-envelope");
            nsmgr.AddNamespace("c", "http://ns.electronichealth.net.au/pcehr/xsd/common/CommonCoreElements/1.0");
            nsmgr.AddNamespace("a", "http://www.w3.org/2005/08/addressing");
            nsmgr.AddNamespace("xds", "urn:ihe:iti:xds-b:2007");
            nsmgr.AddNamespace("lcm", "urn:oasis:names:tc:ebxml-regrep:xsd:lcm:3.0");
            nsmgr.AddNamespace("rim", "urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0");
            XmlNode payload = soapRequest.SelectSingleNode("/s:Envelope/s:Body/xds:ProvideAndRegisterDocumentSetRequest", nsmgr);
            soapRequest.Schemas.Add(nsmgr.LookupNamespace("s"), "soap-envelope.xsd");
            soapRequest.Schemas.Add(nsmgr.LookupNamespace("c"), "PCEHR_CommonTypes.xsd");
            soapRequest.Schemas.Add(nsmgr.LookupNamespace("xds"), "XDS.b_DocumentRepository.xsd");
            soapRequest.Schemas.Add(nsmgr.LookupNamespace("lcm"), "lcm.xsd");
            soapRequest.Schemas.Add(nsmgr.LookupNamespace("rim"), "rim.xsd");
            try
            {
                soapRequest.Validate(null);
                LogAssert.IsTrue(soapRequest.SchemaInfo.Validity == XmlSchemaValidity.Valid,
                    "SOAP request validates against SOAP envelope schema",
                    "SOAP request is invalid");

                soapRequest.Validate(null, payload);
                LogAssert.IsTrue(payload.SchemaInfo.Validity == XmlSchemaValidity.Valid,
                    "ProvideAndRegisterDocumentSetRequest validates against XDS.b Document Repository schema.",
                    "ProvideAndRegisterDocumentSetRequest is invalid");
            }
            catch (XmlException ex)
            {
                LogAssert.Fail(ex.Message);
            }
            catch (XmlSchemaValidationException ex)
            {
                LogAssert.Fail(ex.Message);
            }
            NpdrHeaderTests headerTests = new NpdrHeaderTests(this);
            headerTests.CheckHeader(soapRequest, DialogueResource.HipsServiceUploadDocument);
            EvaluationC(payload, nsmgr);
            EvaluationD(payload, nsmgr);
            EvaluationE(payload, nsmgr);
            EvaluationF(payload, nsmgr);
            EvaluationH(soapRequest, nsmgr);
            EvaluationI(payload, nsmgr);
            EvaluationJ(payload, nsmgr);
            EvaluationK(payload, nsmgr);
            EvaluationL(payload, nsmgr);
            EvaluationM(payload, nsmgr);
            EvaluationN(payload, nsmgr);
            EvaluationO(payload, nsmgr, document);
            EvaluationP(payload, nsmgr);
            EvaluationQ(payload, nsmgr, document, documentFormatCode);
            EvaluationR(payload, nsmgr);
            EvaluationS(payload, nsmgr);
            EvaluationT(payload, nsmgr);
            EvaluationU(payload, nsmgr);
            EvaluationV(payload, nsmgr);
            EvaluationW(payload, nsmgr);
            EvaluationX(payload, nsmgr);

            // Y is skipped because it is the same as U
            EvaluationZ(payload, nsmgr);

            // AA is skipped because it cannot be demonstrated in this message for an original upload.
            EvaluationAB(payload, nsmgr);
        }

        /// <summary>
        /// C. The Provide & Register Document Set transactions contains exactly only one XDS Document element
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationC(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "xds:Document";
            Log("Info: Searching payload for {0}", xPath);
            XmlNodeList documentNodes = payload.SelectNodes(xPath, nsmgr);
            LogAssert.AreEqual(1, documentNodes.Count, "Number of XDS Document elements");
        }

        /// <summary>
        /// D. The Provide & Register Document Set transactions contains exactly one XDSDocumentEntry element
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationD(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject[@objectType='urn:uuid:7edca82f-054d-47f2-a032-9b2a5b5186c1']";
            Log("Info: Searching payload for XDSDocumentEntry elements using XPath {0}", xPath);
            XmlNodeList xdsDocumentEntryNodes = payload.SelectNodes(xPath, nsmgr);
            LogAssert.AreEqual(1, xdsDocumentEntryNodes.Count, "Number of XDSDocumentEntry elements");
        }

        /// <summary>
        /// E. The Provide & Register Document Set transactions DOES NOT contain XDS Folder elements
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationE(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject[@objectType='urn:uuid:d9d542f3-6cc4-48b6-8870-ea235fbc94c2']";
            Log("Info: Searching payload for XDSFolder elements using XPath {0}", xPath);
            XmlNodeList xdsFolderNodes = payload.SelectNodes(xPath, nsmgr);
            LogAssert.AreEqual(0, xdsFolderNodes.Count, "Number of XDS Folder elements");
        }

        /// <summary>
        /// F. Document Addendum, Document Transformation and Folder Management Options are NOT allowed.
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationF(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:Association[@associationType='urn:ihe:iti:2007:AssociationType:APND']";
            Log("Info: Searching payload for Document Addendum associations using XPath {0}", xPath);
            XmlNodeList apndAssociationNodes = payload.SelectNodes(xPath, nsmgr);
            LogAssert.AreEqual(0, apndAssociationNodes.Count, "Number of Document Addendum elements");

            xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:Association[@associationType='urn:ihe:iti:2007:AssociationType:XFRM']";
            Log("Info: Searching payload for Document Transformation associations using XPath {0}", xPath);
            XmlNodeList xfrmAssociationNodes = payload.SelectNodes(xPath, nsmgr);
            LogAssert.AreEqual(0, xfrmAssociationNodes.Count, "Number of Document Transformation elements");
        }

        /// <summary>
        /// Checks H. WS-Action must be urn:ihe:iti:2007:ProvideAndRegisterDocumentSet-b
        /// </summary>
        /// <param name="soapRequest"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationH(XmlDocument soapRequest, XmlNamespaceManager nsmgr)
        {
            string xPath = "/s:Envelope/s:Header/a:Action";
            Log("Info: Searching request for WS-Action using XPath {0}", xPath);
            XmlNode actionNode = soapRequest.SelectSingleNode(xPath, nsmgr);
            LogAssert.AreEqual("urn:ihe:iti:2007:ProvideAndRegisterDocumentSet-b", actionNode.InnerText, "WS-Action");
        }

        /// <summary>
        /// Checks I: Only XDSDocumentEntry fields which are required in the
        /// XDS.b Field Name of Table 2 in the PCEHR Document Exchange Service
        /// TSS are provided.
        /// 
        /// NPDR Supplement: Must also provide hash, size and repositoryUniqueId.
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationI(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string[] shownSlotNames =
            {
                "creationTime",
                "serviceStartTime",
                "serviceStopTime",
                "languageCode",
                "sourcePatientId",
                "hash",
                "size",
                "repositoryUniqueId"
            };
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Slot";
            Log("Info: Searching payload XDSDocumentEntry slots using XPath {0}", xPath);
            XmlNodeList xdsDocumentEntrySlots = payload.SelectNodes(xPath, nsmgr);
            foreach (XmlNode node in xdsDocumentEntrySlots)
            {
                string name = node.Attributes["name"].Value;
                bool validName = shownSlotNames.Contains(name);
                string passMessage = string.Format("XDSDocumentEntry slot {0} is shown in the table.", name);
                string failMessage = string.Format("XDSDocumentEntry slot {0} is not shown in the table", name);
                LogAssert.IsTrue(validName, passMessage, failMessage);
            }

            Dictionary<string, string> requiredXDSDocumentEntryClassificationSchemes = new Dictionary<string, string>()
            {
                {"urn:uuid:93606bcf-9494-43ec-9b4e-a7748d1a838d", "XDSDocumentEntry.author"},
                {"urn:uuid:41a5887f-8865-4c09-adf7-e362475b143a", "XDSDocumentEntry.classCode"},
                {"urn:uuid:a09d5840-386c-46f2-b5ad-9c3699a4309d", "XDSDocumentEntry.formatCode"},
                {"urn:uuid:2e82c1f6-a085-4c72-9da3-8640a32e42ab", "XDSDocumentEntry.uniqueId"},
                {"urn:uuid:f33fb8ac-18af-42cc-ae0e-ed0b0bdb91e1", "XDSDocumentEntry.healthCareFacilityTypeCode"},
                {"urn:uuid:cccf5598-8b07-4b77-a05e-ae952c785ead", "XDSDocumentEntry.practiceSettingCode"},
                {"urn:uuid:f4f85eac-e6cb-4883-b524-f2705394840f", "XDSDocumentEntry.confidentialityCode"},
                {"urn:uuid:f0306f51-975f-434e-a61c-c59651d33983", "XDSDocumentEntry.typeCode"},
                {"urn:uuid:58a6f841-87b3-4a3e-92fd-a8ffeff98427", "XDSDocumentEntry.patientId"},
            };
            xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Classification";
            Log("Info: Searching payload for XDSDocumentEntry classifications using XPath {0}", xPath);
            XmlNodeList xdsDocumentEntryClassifications = payload.SelectNodes(xPath, nsmgr);
            foreach (XmlNode node in xdsDocumentEntryClassifications)
            {
                string classificationScheme = node.Attributes["classificationScheme"].Value;
                bool validClassificationScheme = requiredXDSDocumentEntryClassificationSchemes.ContainsKey(classificationScheme);
                string passMessage = string.Format("XDSDocumentEntry classification {1} with UUID {0} is shown in the table.",
                    classificationScheme, requiredXDSDocumentEntryClassificationSchemes[classificationScheme]);
                string failMessage = string.Format("XDSDocumentEntry classification with UUID {0} is not shown in the table.",
                    classificationScheme);
                LogAssert.IsTrue(validClassificationScheme, passMessage, failMessage);
            }
        }

        /// <summary>
        /// Check J: The XDSDocumentEntry.sourcePatientId contains 16-digit IHI number in the PCEHR Header.
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationJ(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "/s:Envelope/s:Header/c:PCEHRHeader/c:ihiNumber";
            Log("Info: Searching SOAP request for ihiNumber using XPath {0}", xPath);
            XmlNode ihiNumberNode = payload.SelectSingleNode(xPath, nsmgr);
            xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Slot[@name='sourcePatientId']/rim:ValueList/rim:Value";
            Log("Info: Searching payload for XDSDocumentEntry.sourcePatientId using XPath {0}", xPath);
            XmlNode sourcePatientIdNode = payload.SelectSingleNode(xPath, nsmgr);
            bool sourcePatientIdContainsIhiNumber = sourcePatientIdNode.InnerText.Contains(ihiNumberNode.InnerText);
            string passMessage = string.Format("Source Patient ID {0} contains IHI number {1}", sourcePatientIdNode.InnerText, ihiNumberNode.InnerText);
            string failMessage = string.Format("Source Patient ID {0} does not contain IHI number {1}", sourcePatientIdNode.InnerText, ihiNumberNode.InnerText);
            LogAssert.IsTrue(sourcePatientIdContainsIhiNumber, passMessage, failMessage);
        }

        /// <summary>
        /// Check K. The XDSDocumentEntry.confidentialityCode contains 'N/A'
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationK(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Classification[@classificationScheme='urn:uuid:f4f85eac-e6cb-4883-b524-f2705394840f']/rim:Name/rim:LocalizedString";
            Log("Info: Searching payload for XDSDocumentEntry.confidentialityCode using XPath {0}", xPath);
            XmlNode confidentialityCodeNode = payload.SelectSingleNode(xPath, nsmgr);
            LogAssert.AreEqual("N/A", confidentialityCodeNode.Attributes["value"].Value, "Confidentiality Code Classification Name LocalizedString value");
        }

        /// <summary>
        /// Check L. The XDSDocumentEntry.uniqueId contains a GUID or an OID.
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationL(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:ExternalIdentifier[@identificationScheme='urn:uuid:2e82c1f6-a085-4c72-9da3-8640a32e42ab']";
            Log("Info: Searching payload for XDSDocumentEntry.uniqueId using XPath {0}", xPath);
            XmlNode uniqueIdNode = payload.SelectSingleNode(xPath, nsmgr);
            string uniqueId = uniqueIdNode.Attributes["value"].Value;
            bool isOid = Regex.IsMatch(uniqueId, @"^\d+(\.\d+)*(\^.*)?$");
            LogAssert.IsTrue(isOid,
                string.Format("Unique ID {0} is an OID", uniqueId),
                string.Format("Unique ID {0} is not an OID", uniqueId));
        }

        /// <summary>
        /// Check M. The XDSDocumentEntry.typeCode is the same value as the XDSDocumentEntry.classCode field
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationM(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Classification[@classificationScheme='urn:uuid:41a5887f-8865-4c09-adf7-e362475b143a']";
            Log("Info: Searching payload for XDSDocumentEntry.classCode using XPath {0}", xPath);
            XmlNode classCodeNode = payload.SelectSingleNode(xPath, nsmgr);
            string classCode = classCodeNode.Attributes["nodeRepresentation"].Value;
            xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Classification[@classificationScheme='urn:uuid:f0306f51-975f-434e-a61c-c59651d33983']";
            Log("Info: Searching payload for XDSDocumentEntry.typeCode using XPath {0}", xPath);
            XmlNode typeCodeNode = payload.SelectSingleNode(xPath, nsmgr);
            string typeCode = typeCodeNode.Attributes["nodeRepresentation"].Value;

            LogAssert.AreEqual(classCode, typeCode, "XDSDocumentEntry.classCode", "XDSDocumentEntry.typeCode");
        }

        /// <summary>
        /// PCEHR Document Exchange Service IHE XDS-b Technical Service Specification v1.4 12/04/2013
        /// Table 3 - XDSDocumentEntry Document Type and Class Code value set.
        /// </summary>
        private class Table3
        {
            private static List<Table3> allEntries;

            public string CodingSystem { get; set; }

            public string Code { get; set; }

            public string ClassCodeDisplayName { get; set; }

            public string TypeCodeDisplayName { get; set; }

            private Table3(string codingSystem, string code, string classCodeDisplayName, string typeCodeDisplayName)
            {
                CodingSystem = codingSystem;
                Code = code;
                ClassCodeDisplayName = classCodeDisplayName;
                TypeCodeDisplayName = typeCodeDisplayName;
            }

            /// <summary>
            /// Gets the entries in the table indexed by their code.
            /// </summary>
            public static Dictionary<string, Table3> ByCode
            {
                get
                {
                    return AllEntries.ToDictionary(a => a.Code);
                }
            }

            /// <summary>
            /// Gets all the entries in the table.
            /// </summary>
            public static List<Table3> AllEntries
            {
                get
                {
                    if (allEntries == null)
                    {
                        allEntries = new List<Table3>();
                        allEntries.Add(new Table3("LOINC", "60591-5", "Patient Summary", "Shared Health Summary"));
                        allEntries.Add(new Table3("LOINC", "57133-1", "Referral note", "e-Referral"));
                        allEntries.Add(new Table3("LOINC", "51852-2", "Letter", "Specialist Letter"));

                        // Note from PCEHR SVT Support:
                        // As per HL7 terminologies it is supposed to be "Discharge Summarization Note" with a z.
                        // However, if you use the vendor library and that is generating the incorrect spelling we’ll consider it a pass.
                        //allEntries.Add(new Table3("LOINC", "18842-5", "Discharge Summarization Note", "Discharge Summary"));
                        allEntries.Add(new Table3("LOINC", "18842-5", "Discharge Summarisation Note", "Discharge Summary"));

                        allEntries.Add(new Table3("LOINC", "34133-9", "Summarization of episode note", "Event Summary"));
                        allEntries.Add(new Table3("NCTIS", "100.161", "e-Prescription", "e-Prescription"));
                        allEntries.Add(new Table3("NCTIS", "100.16112", "Dispense Record", "Dispense Record"));
                        allEntries.Add(new Table3("NCTIS", "100.16285", "Prescription Request", "Prescription Request"));
                        allEntries.Add(new Table3("NCTIS", "100.1665", "Pharmaceutical Benefits Report", "Pharmaceutical Benefits Report"));
                        allEntries.Add(new Table3("NCTIS", "100.16659", "Australian Childhood Immunisation Register", "Australian Childhood Immunisation Register"));
                        allEntries.Add(new Table3("NCTIS", "100.16644", "Medicare/DVA Benefits Report", "Medicare/DVA Benefits Report"));
                        allEntries.Add(new Table3("NCTIS", "102.16671", "Australian Organ Donor Register", "Australian Organ Donor Register"));
                        allEntries.Add(new Table3("NCTIS", "100.16681", "Consumer Entered Notes", "Consumer Entered Notes"));
                        allEntries.Add(new Table3("NCTIS", "100.16685", "Consumer Entered Health Summary", "Consumer Entered Health Summary"));
                        allEntries.Add(new Table3("NCTIS", "100.16696", "Advanced Care Directive Custodian Record", "Advanced Care Directive Custodian Record"));
                        allEntries.Add(new Table3("NCTIS", "100.16764", "PCEHR Prescription Record", "PCEHR Prescription Record"));
                        allEntries.Add(new Table3("NCTIS", "100.16765", "PCEHR Dispense Record", "PCEHR Dispense Record"));
                    }
                    return allEntries;
                }
            }
        }

        /// <summary>
        /// Check N. The XDSDocumentEntry.typeCodeDisplayName is set to the appropriate typeCodeDisplayName
        /// and the XDSDocumentEntry.classCodeDisplayName is set to the appropriate classCodeDisplayName
        ///
        /// Note: Accenture advised the above text on 10/07/2013 as a replacement for the incorrect text in NOC Test Plan 4.0:
        /// "N. The XDSDocumentEntry.typeCodeDisplayName is the same value as the XDSDocumentEntry.classCodeDisplayName field"
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationN(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Classification[@classificationScheme='urn:uuid:41a5887f-8865-4c09-adf7-e362475b143a']";
            Log("Info: Searching payload for XDSDocumentEntry.classCode using XPath {0}", xPath);
            XmlNode classCodeNode = payload.SelectSingleNode(xPath, nsmgr);
            string classCode = classCodeNode.Attributes["nodeRepresentation"].Value;
            xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Classification[@classificationScheme='urn:uuid:f0306f51-975f-434e-a61c-c59651d33983']";
            Log("Info: Searching payload for XDSDocumentEntry.typeCode using XPath {0}", xPath);
            XmlNode typeCodeNode = payload.SelectSingleNode(xPath, nsmgr);
            string typeCode = typeCodeNode.Attributes["nodeRepresentation"].Value;
            xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Classification[@classificationScheme='urn:uuid:41a5887f-8865-4c09-adf7-e362475b143a']/rim:Name/rim:LocalizedString";
            Log("Info: Searching payload for XDSDocumentEntry.classCodeDisplayName using XPath {0}", xPath);
            XmlNode classCodeDisplayNameNode = payload.SelectSingleNode(xPath, nsmgr);
            string classCodeDisplayName = classCodeDisplayNameNode.Attributes["value"].Value;
            xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Classification[@classificationScheme='urn:uuid:f0306f51-975f-434e-a61c-c59651d33983']/rim:Name/rim:LocalizedString";
            Log("Info: Searching payload for XDSDocumentEntry.typeCodeDisplayName using XPath for {0}", xPath);
            XmlNode typeCodeDisplayNameNode = payload.SelectSingleNode(xPath, nsmgr);
            string typeCodeDisplayName = typeCodeDisplayNameNode.Attributes["value"].Value;

            LogAssert.AreEqual(Table3.ByCode[classCode].ClassCodeDisplayName, classCodeDisplayName, "XDSDocumentEntry.classCodeDisplayName");
            LogAssert.AreEqual(Table3.ByCode[typeCode].TypeCodeDisplayName, typeCodeDisplayName, "XDSDocumentEntry.typeCodeDisplayName");
        }

        /// <summary>
        /// Check O. The XDSDocumentEntry.uniqueId is identical to the DocumentID field within the CDA document
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationO(XmlNode payload, XmlNamespaceManager nsmgr, CdaDocument document)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:ExternalIdentifier[@identificationScheme='urn:uuid:2e82c1f6-a085-4c72-9da3-8640a32e42ab']";
            Log("Info: Searching payload for XDSDocumentEntry.uniqueId using XPath {0}", xPath);
            XmlNode uniqueIdNode = payload.SelectSingleNode(xPath, nsmgr);
            string uniqueId = uniqueIdNode.Attributes["value"].Value;
            string cdaDocId = document.GetDocId();
            bool isOid = Regex.IsMatch(cdaDocId, @"^\d+(\.\d+)*$");
            if (isOid)
            {
                LogAssert.AreEqual(cdaDocId, uniqueId, "the DocumentID field within the CDA document", "The XDSDocumentEntry.uniqueId");
            }
            else
            {
                LogAssert.AreEqual(CdaDocument.ConvertToOid(cdaDocId), uniqueId, "the DocumentID field within the CDA document (converted to an OID)", "The XDSDocumentEntry.uniqueId");
            }
        }

        /// <summary>
        /// Check P. The XDSDocumentEntry.sourcePatientId is identical to the XDSDocumentEntry.patientId field within the supplied CDA Document
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationP(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:ExternalIdentifier[@identificationScheme='urn:uuid:58a6f841-87b3-4a3e-92fd-a8ffeff98427']";
            Log("Info: Searching payload for XDSDocumentEntry.patientId using XPath {0}", xPath);
            XmlNode patientIdNode = payload.SelectSingleNode(xPath, nsmgr);
            string patientId = patientIdNode.Attributes["value"].Value;

            xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Slot[@name='sourcePatientId']/rim:ValueList/rim:Value";
            Log("Info: Searching payload for XDSDocumentEntry.sourcePatientId using XPath {0}", xPath);
            XmlNode sourcePatientIdNode = payload.SelectSingleNode(xPath, nsmgr);
            string sourcePatientId = sourcePatientIdNode.InnerText;

            LogAssert.AreEqual(patientId, sourcePatientId, "XDSDocumentEntry.patientId", "XDSDocumentEntry.sourcePatientId");
        }

        /// <summary>
        /// Check Q. The XDSDocumentEntry.formatCode field contains the Template ID of the Template to which the CDA Document asserts conformance
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationQ(XmlNode payload, XmlNamespaceManager nsmgr, CdaDocument document, string documentFormatCode)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Classification[@classificationScheme='urn:uuid:a09d5840-386c-46f2-b5ad-9c3699a4309d']";
            Log("Info: Searching payload for XDSDocumentEntry.formatCode using XPath {0}", xPath);
            XmlNode formatCodeNode = payload.SelectSingleNode(xPath, nsmgr);
            string formatCode = formatCodeNode.Attributes["nodeRepresentation"].Value;

            LogAssert.AreEqual(documentFormatCode, formatCode, "XDSDocumentEntry.formatCode");
        }

        /// <summary>
        /// Check R. The XDSDocumentEntry.languageCode field is set to a fixed value of 'en-AU'
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationR(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Slot[@name='languageCode']/rim:ValueList/rim:Value";
            Log("Info: Searching payload for XDSDocumentEntry.languageCode using XPath {0}", xPath);
            XmlNode languageCodeNode = payload.SelectSingleNode(xPath, nsmgr);
            string languageCode = languageCodeNode.InnerText;

            LogAssert.AreEqual("en-AU", languageCode, "XDSDocumentEntry.languageCode");
        }

        /// <summary>
        /// Check S: Only XDSSubmissionSet fields which are shown in the XDS.b
        /// Field Name of Table 3 in the  PCEHR Document Exchange Service TSS
        /// are provided.
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationS(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string[] requiredXDSSubmissionSetSlotNames =
            {
                "submissionTime",
            };
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:RegistryPackage/rim:Slot";
            Log("Info: Searching payload for XDSSubmissionSet slots using XPath {0}", xPath);
            XmlNodeList xdsSubmissionSetSlots = payload.SelectNodes(xPath, nsmgr);
            foreach (XmlNode node in xdsSubmissionSetSlots)
            {
                string name = node.Attributes["name"].Value;
                bool validName = requiredXDSSubmissionSetSlotNames.Contains(name);
                string passMessage = string.Format("XDSSubmissionSet slot {0} is shown in the table.", name);
                string failMessage = string.Format("XDSSubmissionSet slot {0} is not shown in the table", name);
                LogAssert.IsTrue(validName, passMessage, failMessage);
            }

            Dictionary<string, string> requiredXDSSubmissionSetClassificationSchemes = new Dictionary<string, string>()
            {
                {"urn:uuid:a7058bb9-b4e4-4307-ba5b-e3f0ab85e12d", "XDSSubmissionSet.author"},
                {"urn:uuid:aa543740-bdda-424e-8c96-df4873be8500", "XDSSubmissionSet.contentTypeCode"},
                {"urn:uuid:96fdda7c-d067-4183-912e-bf5ee74998a8", "XDSSubmissionSet.uniqueId"},
                {"urn:uuid:554ac39e-e3fe-47fe-b233-965d2a147832", "XDSSubmissionSet.sourceId"},
                {"urn:uuid:6b5aea1a-874d-4603-a4bc-96a0a7b38446", "XDSSubmissionSet.patientId"}
            };

            xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:RegistryPackage/rim:Classification";
            Log("Info: Searching payload for XDSSubmissionSet Classifications using XPath {0}", xPath);
            XmlNodeList xdsSubmissionSetClassifications = payload.SelectNodes(xPath, nsmgr);
            foreach (XmlNode node in xdsSubmissionSetClassifications)
            {
                string classificationScheme = node.Attributes["classificationScheme"].Value;
                bool validClassificationScheme = requiredXDSSubmissionSetClassificationSchemes.ContainsKey(classificationScheme);
                string passMessage = string.Format("XDSSubmissionSet classification {1} with UUID {0} is shown in the table.",
                    classificationScheme, requiredXDSSubmissionSetClassificationSchemes[classificationScheme]);
                string failMessage = string.Format("XDSSubmissionSet classification with UUID {0} is not shown in the table.",
                    classificationScheme);
                LogAssert.IsTrue(validClassificationScheme, passMessage, failMessage);
            }
        }

        /// <summary>
        /// Check T. The XDSSubmissionSet.entryUUID contains a symbolic identifier of 'SUBSET_SYMBOLICID'
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationT(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:RegistryPackage";
            Log("Info: Searching payload for XDSSubmissionSet.entryUUID using XPath {0}", xPath);
            XmlNode entryUUIDNode = payload.SelectSingleNode(xPath, nsmgr);
            string entryUUID = entryUUIDNode.Attributes["id"].Value;
            LogAssert.AreEqual("SUBSET_SYMBOLICID_01", entryUUID, "XDSSubmissionSet.entryUUID");
        }

        /// <summary>
        /// Check U. The XDSSubmissionSet.authorPerson entity is identical to the value provided within the XDSDocumentEntry.authorPerson
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationU(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:RegistryPackage/rim:Classification[@classificationScheme='urn:uuid:a7058bb9-b4e4-4307-ba5b-e3f0ab85e12d']/rim:Slot[@name='authorPerson']/rim:ValueList/rim:Value";
            Log("Info: Searching payload for XDSSubmissionSet.authorPerson using XPath {0}", xPath);
            XmlNode submissionSetAuthorPersonNode = payload.SelectSingleNode(xPath, nsmgr);
            string submissionSetAuthorPerson = submissionSetAuthorPersonNode.InnerText;

            xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Classification[@classificationScheme='urn:uuid:93606bcf-9494-43ec-9b4e-a7748d1a838d']/rim:Slot[@name='authorPerson']/rim:ValueList/rim:Value";
            Log("Info: Searching payload for XDSDocumentEntry.authorPerson using XPath {0}", xPath);
            XmlNode documentEntryAuthorPersonNode = payload.SelectSingleNode(xPath, nsmgr);
            string documentEntryAuthorPerson = documentEntryAuthorPersonNode.InnerText;

            LogAssert.AreEqual(documentEntryAuthorPerson, submissionSetAuthorPerson, "XDSDocumentEntry.authorPerson", "XDSSubmissionSet.authorPerson");
        }

        /// <summary>
        /// Check V. The XDSSubmissionSet.authorInstitution entity is identical to the value provided within the XDSDocumentEntry.authorInstitution
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationV(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:RegistryPackage/rim:Classification[@classificationScheme='urn:uuid:a7058bb9-b4e4-4307-ba5b-e3f0ab85e12d']/rim:Slot[@name='authorInstitution']/rim:ValueList/rim:Value";
            Log("Info: Searching payload for XDSSubmissionSet.authorInstitution using XPath {0}", xPath);
            XmlNode submissionSetAuthorInstitutionNode = payload.SelectSingleNode(xPath, nsmgr);
            string submissionSetAuthorInstitution = submissionSetAuthorInstitutionNode.InnerText;

            xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Classification[@classificationScheme='urn:uuid:93606bcf-9494-43ec-9b4e-a7748d1a838d']/rim:Slot[@name='authorInstitution']/rim:ValueList/rim:Value";
            Log("Info: Searching payload for XDSDocumentEntry.authorInstitution using XPath {0}", xPath);
            XmlNode documentEntryAuthorInstitutionNode = payload.SelectSingleNode(xPath, nsmgr);
            string documentEntryAuthorInstitution = documentEntryAuthorInstitutionNode.InnerText;

            LogAssert.AreEqual(documentEntryAuthorInstitution, submissionSetAuthorInstitution, "XDSDocumentEntry.authorInstitution", "XDSSubmissionSet.authorInstitution");
        }

        /// <summary>
        /// Check W. The XDSSubmission.contentTypeCode entity is identical to the value in the XDSDocumentEntry.classCode entity
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationW(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:RegistryPackage/rim:Classification[@classificationScheme='urn:uuid:aa543740-bdda-424e-8c96-df4873be8500']";
            Log("Info: Searching payload for XDSSubmissionSet.contentTypeCode using XPath {0}", xPath);
            string contentTypeCode = payload.SelectSingleNode(xPath, nsmgr).Attributes["nodeRepresentation"].Value;

            xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Classification[@classificationScheme='urn:uuid:41a5887f-8865-4c09-adf7-e362475b143a']";
            Log("Info: Searching payload for XDSDocumentEntry.classCode using XPath {0}", xPath);
            string classCode = payload.SelectSingleNode(xPath, nsmgr).Attributes["nodeRepresentation"].Value;

            LogAssert.AreEqual(classCode, contentTypeCode, "XDSDocumentEntry.classCode", "XDSSubmissionSet.contentTypeCode");
        }

        /// <summary>
        /// Check X. The XDSSubmissionSet.contentTypeCodeDisplayName entity is identical with the value provided within the XDSDocumentEntry.classCodeDisplayName entity
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationX(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:RegistryPackage/rim:Classification[@classificationScheme='urn:uuid:aa543740-bdda-424e-8c96-df4873be8500']/rim:Name/rim:LocalizedString";
            Log("Info: Searching payload for XDSSubmissionSet.contentTypeCodeDisplayName using XPath {0}", xPath);
            string contentTypeCodeDisplayName = payload.SelectSingleNode(xPath, nsmgr).InnerText;

            xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Classification[@classificationScheme='urn:uuid:41a5887f-8865-4c09-adf7-e362475b143a']/rim:Name/rim:LocalizedString";
            Log("Info: Searching payload for XDSDocumentEntry.classCodeDisplayName using XPath {0}", xPath);
            string classCodeDisplayName = payload.SelectSingleNode(xPath, nsmgr).InnerText;

            LogAssert.AreEqual(classCodeDisplayName, contentTypeCodeDisplayName, "XDSDocumentEntry.classCodeDisplayName", "XDSSubmissionSet.contentTypeCodeDisplayName");
        }

        /// <summary>
        /// Check Z. The XDSSubmissionSet.personID entity is identical to the value provided within the XDSDocumentEntry.sourcePersonId entity
        /// (actually patientID not personID, and sourcePatientID not sourcePersonID)
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationZ(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:RegistryPackage/rim:ExternalIdentifier[@identificationScheme='urn:uuid:6b5aea1a-874d-4603-a4bc-96a0a7b38446']";
            Log("Info: Searching payload for XDSSubmissionSet.patientID using XPath {0}", xPath);
            string patientID = payload.SelectSingleNode(xPath, nsmgr).Attributes["value"].Value;

            xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:ExtrinsicObject/rim:Slot[@name='sourcePatientId']/rim:ValueList/rim:Value";
            Log("Info: Searching payload for XDSDocumentEntry.sourcePatientId using XPath {0}", xPath);
            string sourcePatientId = payload.SelectSingleNode(xPath, nsmgr).InnerText;

            LogAssert.AreEqual(sourcePatientId, patientID, "XDSDocumentEntry.sourcePatientId", "XDSSubmissionSet.patientID");
        }

        /// <summary>
        /// Check AB. The XDS.b Association Type is set to 'urn:ihe:ihi:2007:AssociationType:HasMember'
        /// Note: This test is incorrect, the Association Type should be 'urn:oasis:names:tc:ebxml-regrep:AssociationType:HasMember'.
        /// </summary>
        /// <param name="payload"></param>
        /// <param name="nsmgr"></param>
        private void EvaluationAB(XmlNode payload, XmlNamespaceManager nsmgr)
        {
            string xPath = "lcm:SubmitObjectsRequest/rim:RegistryObjectList/rim:Association";
            Log("Info: Searching payload for Association using XPath {0}", xPath);
            string associationType = payload.SelectSingleNode(xPath, nsmgr).Attributes["associationType"].Value;

            LogAssert.AreEqual("urn:oasis:names:tc:ebxml-regrep:AssociationType:HasMember", associationType, "XDS.b Association Type");
        }

        /// <summary>
        /// Test No:         NOC_DC_007
        /// 
        /// Objective/Input:
        /// Perform a ProvideAndRegisterDocumentSet-b operation to amend a
        /// document which does not exist in the NPDR system.
        ///
        /// Expected Result:
        /// Document should fail submission as the document ID and version for
        /// which the Amend is being submitted for does not exist within the
        /// NPDR.
        /// </summary>
        [TestMethod]
        [TestCategory("NPDR_NOC")]
        public void NPDR_NOC_07()
        {
            LogAssert.IsTrue(true, "HIPS does not allow the amendment of a document which does not exist in the NPDR system.", "no");
        }

        /// <summary>
        /// Test No:         NOC_DC_003
        /// Objective/Input:
        /// To show that the client system is able to replace (amend) a
        /// prescribe record that exists in the NPDR System. 
        ///
        /// Expected Result:
        /// Original document will be replaced and registered with the latest
        /// amended version and be successfully registered with the PCEHR.
        /// </summary>
        [TestMethod]
        [TestCategory("NPDR_NOC")]
        public void NPDR_NOC_03()
        {
            TestSuccessfulSupersede(SampleDocumentType.PcehrPrescriptionRecord);
        }

        /// <summary>
        /// Test No:         NOC_DC_004
        /// Objective/Input:
        /// To show that the client system is able to replace (amend) a
        /// dispense record that exists in the NPDR System. 
        ///
        /// Expected Result:
        /// Original document will be replaced and registered with the latest
        /// amended version and be successfully registered with the PCEHR.
        /// </summary>
        [TestMethod]
        [TestCategory("NPDR_NOC")]
        public void NPDR_NOC_04()
        {
            TestSuccessfulSupersede(SampleDocumentType.PcehrDispenseRecord);
        }

        /// <summary>
        /// Tests the successful supersede of a document on NPDR.
        /// </summary>
        /// <param name="docType"></param>
        private void TestSuccessfulSupersede(SampleDocumentType docType)
        {
            patient = CcaPatient.GetPatient(ihiValid: true, validatedWithinPeriod: true, hasUnresolvedAlerts: false, testDataId: "NOC ID 17");
            UserDetails user = patient.GetTestUser();
            PatientIdentifierBase identifier = patient.TargetPatientIdentifier;
            CdaDocument document = patient.CreateNewDocument(docType);
            QueueHelper.UploadDocumentAndWaitUntilProcessed(patient, LogAssert, document);

            CdaDocument document2 = patient.CreateAmendedDocument(document);
            QueueHelper.UploadDocumentAndWaitUntilProcessed(patient, LogAssert, document2);
        }

    }
}